from abc import ABC, abstractmethod
from qec_lego_bench.codes.code import Code
import stim


class StabilizerCode(Code, ABC):

    @property
    @abstractmethod
    def n(self) -> int: ...

    @property
    @abstractmethod
    def stabilizers(self) -> list[stim.PauliString]: ...

    def init_logical_operators(self):
        # https://quantumcomputing.stackexchange.com/questions/37812/how-to-find-a-set-of-independent-logical-operators-for-a-stabilizer-code-with-st
        completed_tableau = stim.Tableau.from_stabilizers(
            self.stabilizers,
            allow_redundant=True,
            allow_underconstrained=True,
        )
        self._logical_operators = []
        for k in range(len(completed_tableau))[::-1]:
            z = completed_tableau.z_output(k)
            if z in self.stabilizers:
                break
            x = completed_tableau.x_output(k)
            self._logical_operators.append((x, z))

    def init_circuit(self):
        c = stim.Circuit()
        # add qubit coordinates
        for qubit_index, coordinate in enumerate(self.qubit_coordinates):
            if coordinate is not None:
                c.append("QUBIT_COORDS", qubit_index, coordinate)
        # prepare the qubits in some random stabilizer states and logical state
        for stabilizer in self.stabilizers:
            assert len(stabilizer) <= self.n
            c.append("MPP", stim.target_combined_paulis(stabilizer))
        for logical_x, logical_z in self.logical_operators:
            assert len(logical_x) <= self.n
            assert len(logical_z) <= self.n
            logical = logical_x if self.prepare_logical_x_instead_of_z else logical_z
            c.append("MPP", stim.target_combined_paulis(logical))
        bias = len(self.stabilizers) + len(self.logical_operators)
        # add errors at TICK positions
        c.append("TICK")
        c.append("SHIFT_COORDS", (), (0, 0, 1))
        # add the detectors by measuring the stabilizers again and compare the difference
        for stabilizer, coordinates in zip(
            self.stabilizers, self.stabilizer_coordinates
        ):
            c.append("MPP", stim.target_combined_paulis(stabilizer))
            c.append(
                "DETECTOR",
                [stim.target_rec(-1), stim.target_rec(-1 - bias)],
                coordinates,
            )
        # check logical errors by measuring the logical operators
        for logical_index, (logical_x, logical_z) in enumerate(self.logical_operators):
            logical = logical_x if self.prepare_logical_x_instead_of_z else logical_z
            c.append("MPP", stim.target_combined_paulis(logical))
            c.append(
                "OBSERVABLE_INCLUDE",
                [stim.target_rec(-1), stim.target_rec(-1 - bias)],
                logical_index,
            )
        self._circuit = c

    def __init__(self):
        # construct logical observables to check logical errors
        self.init_logical_operators()
        # construct the circuit of the code
        self.init_circuit()
        # call the parent constructor
        super().__init__()

    @property
    def logical_operators(self) -> list[tuple[stim.PauliString]]:
        return self._logical_operators

    @property
    def circuit(self) -> stim.Circuit:
        return self._circuit

    @property
    def qubit_coordinates(self) -> list[tuple[float, float, float] | None]:
        return [None] * self.n

    @property
    def stabilizer_coordinates(self) -> list[tuple[float, float, float] | None]:
        return [None] * len(self.stabilizers)

    @property
    def prepare_logical_x_instead_of_z(self) -> bool:
        return False



from qec_lego_bench.cli.decoders import decoder_cli
from sinter._decoding._decoding_pymatching import PyMatchingDecoder
from sinter._decoding._decoding_fusion_blossom import FusionBlossomDecoder
from mwpf import SinterMWPFDecoder, SinterHUFDecoder, SinterSingleHairDecoder


decoder_cli("PyMatchingDecoder", "pymatching", "mwpm")(PyMatchingDecoder)
decoder_cli("FusionBlossomDecoder", "fusion_blossom", "fb")(FusionBlossomDecoder)
decoder_cli("HUF")(SinterHUFDecoder)
decoder_cli("MWPF", "mw_parity_factor")(SinterMWPFDecoder)
decoder_cli("MWPF_SH")(SinterSingleHairDecoder)



from .noise import Noise
from dataclasses import dataclass
from qec_lego_bench.cli.noises import noise_cli
import stim


@noise_cli("NoNoise")
@dataclass
class NoNoise(Noise):
    """
    no noise
    """

    def __call__(self, circuit: stim.Circuit) -> stim.Circuit:
        return circuit.copy()



from stim import Circuit
from .code import Code
from qec_lego_bench.cli.codes import code_cli


@code_cli("CircuitCode", "file")
class CircuitCode(Code):
    """
    A code directly from a circuit in a file
    """

    def __init__(self, filepath: str):
        self._circuit = Circuit.from_file(filepath)
        super().__init__()

    @property
    def circuit(self) -> Circuit:
        return self._circuit



from .util import named_kwargs_of, params_of_func_or_cls
import sys

registered_decoder_names = {}


def decoder_cli(*decoder_names: str):
    def decorator(constructor):
        params = params_of_func_or_cls(constructor)
        for decoder_name in decoder_names:
            assert decoder_name.isidentifier()
            decoder_name = decoder_name.lower()
            if decoder_name in registered_decoder_names:
                print(
                    f"[warning] decoder name {decoder_name} already registered",
                    file=sys.stderr,
                )
            registered_decoder_names[decoder_name] = (constructor, params)
        return constructor

    return decorator


class DecoderCli:
    def __init__(self, input: str):
        try:
            self.input = input
            decoder_name, params = named_kwargs_of(input)
            decoder_name = decoder_name.lower()
            if decoder_name not in registered_decoder_names:
                print(
                    f"[error] decoder name '{decoder_name}' not found, possible values: {', '.join(registered_decoder_names.keys())}",
                    file=sys.stderr,
                )
                raise ValueError()
            cls, expected_params = registered_decoder_names[decoder_name]
            kwargs = {}
            for param in params:
                assert (
                    param in expected_params
                ), f"unexpected parameter '{param}', expecting one of {', '.join(expected_params.keys())}"
                constructor = expected_params[param]
                kwargs[param] = constructor(params[param])
            self.decoder = cls(**kwargs)
        except Exception as e:
            print(f"[error] {e}", file=sys.stderr)
            raise e

    def __str__(self):
        return self.input

    def __call__(self):
        return self.decoder



from dataclasses import dataclass
from qec_lego_bench.codes.code import Code
from qec_lego_bench.cli.codes import code_cli
import stim
from typing import Optional


@code_cli("NoisyRepetitionCode", "noisy_rep", "rep")
@dataclass
class NoisyRepetitionCode(Code):
    d: int
    p: float = 0.0
    # if not specified, rounds = distance
    rounds: Optional[int] = None

    def __post_init__(self):
        assert self.p >= 0
        assert self.p <= 1
        rounds = self.d if self.rounds is None else self.rounds
        self._circuit = stim.Circuit.generated(
            "repetition_code:memory",
            rounds=rounds,
            distance=self.d,
            after_clifford_depolarization=self.p,
            before_round_data_depolarization=self.p,
            before_measure_flip_probability=self.p,
            after_reset_flip_probability=self.p,
        )

    @property
    def circuit(self) -> stim.Circuit:
        return self._circuit



from typing import Any
import typing
from urllib.parse import parse_qs
import inspect
import types


def kwargs_of_qs(qs: str) -> dict[str, str]:
    """
    parsing a custom querystring format into a dictionary.
    The format is normal querystring format, but with '&' replaced by ','.
    Thus, the value cannot contain ',', which is usually fine in the context of this project.
    """
    querystring = qs.replace(",", "&")
    dict_of_qs = parse_qs(querystring)
    assert len(dict_of_qs) > 0 or qs == "", f"querystring '{qs}' is not valid"
    for key in dict_of_qs:
        assert key.isidentifier(), f"key '{key}' is not a valid identifier"
    return {key: value[0] for key, value in dict_of_qs.items()}


def named_kwargs_of(input: str) -> tuple[str, dict[str, str]]:
    """
    expecting a format of 'name(a=1,b=2)'
    """
    if "(" in input:
        assert input[-1] == ")"
        split_index = input.index("(")
        name = input[:split_index]
        assert name.isidentifier(), f"name '{name}' is not a valid identifier"
        kwargs = kwargs_of_qs(input[split_index + 1 : -1])
        return name, kwargs
    name = input
    assert name.isidentifier(), f"name '{name}' is not a valid identifier"
    return name, {}


def params_of_func_or_cls(func: Any) -> dict[str, Any]:
    """
    the decorated class must have an initialization function that accepts str, int or float KEYWORD input.
    or it could be a function that accepts str, int or float KEYWORD input.
    All other types of arguments must be convertible from str, i.e., cls(str) must work
    """
    signature = inspect.signature(func)
    params = {}
    for param in list(signature.parameters.values()):
        if (
            isinstance(param.annotation, types.UnionType)
            or typing.get_origin(param.annotation) == typing.Union
        ):
            args = [arg for arg in param.annotation.__args__ if arg != type(None)]
            assert len(args) == 1, "only support Union[TYPE, None] for now"
            assert (
                param.default is None
            ), f"default value of {param.name} must be None for Union[TYPE, None] in {func.__name__}"
            params[param.name] = args[0]
        else:
            params[param.name] = param.annotation
    return params



from qec_lego_bench.codes.stabilizer_codes import StabilizerCode
from abc import ABC, abstractmethod
from galois.typing import ArrayLike
import stim
from galois import GF2
import numpy as np


class CSSCode(StabilizerCode, ABC):

    @property
    @abstractmethod
    def H_X(self) -> ArrayLike: ...

    @property
    @abstractmethod
    def H_Z(self) -> ArrayLike: ...

    def css_init_stabilizers(self):
        assert (self.H_X @ self.H_Z.T == 0).all(), "X and Z checks must commute"
        self._stabilizers = []
        for t, matrix in [("X", self.H_X), ("Z", self.H_Z)]:
            for line in matrix:
                self._stabilizers.append(
                    stim.PauliString([t if e == 1 else "I" for e in line])
                )

    @property
    def stabilizers(self) -> list[stim.PauliString]:
        return self._stabilizers

    def __init__(self):
        self.css_init_stabilizers()
        super().__init__()

        self.L_X, self.L_Z = construct_logical_checks(self.logical_operators)
        assert (self.H_X @ self.L_Z.T == 0).all()
        assert (self.H_Z @ self.L_X.T == 0).all()


def construct_logical_checks(
    logical_operators: list[tuple[stim.PauliString, stim.PauliString]],
) -> tuple[ArrayLike, ArrayLike]:
    height = len(logical_operators)
    width = len(logical_operators[0][0])
    L_X = GF2(np.zeros((height, width), dtype=np.uint8))
    L_Z = GF2(np.zeros((height, width), dtype=np.uint8))
    for idx, (l_x, l_z) in enumerate(logical_operators):
        for t, l, L in [("X", l_x, L_X), ("Z", l_z, L_Z)]:
            indices = l.pauli_indices(t)
            assert indices == l.pauli_indices(), "should not include other Pauli types"
            L[idx][indices] = 1
    return L_X, L_Z



from .util import named_kwargs_of, params_of_func_or_cls
import sys

registered_code_names = {}


def code_cli(*code_names: str):
    def decorator(constructor):
        params = params_of_func_or_cls(constructor)
        for code_name in code_names:
            assert code_name.isidentifier()
            code_name = code_name.lower()
            if code_name in registered_code_names:
                print(
                    f"[warning] code name {code_name} already registered",
                    file=sys.stderr,
                )
            registered_code_names[code_name] = (constructor, params)
        return constructor

    return decorator


class CodeCli:
    def __init__(self, input: str):
        try:
            self.input = input
            code_name, params = named_kwargs_of(input)
            code_name = code_name.lower()
            if code_name not in registered_code_names:
                print(
                    f"[error] code name '{code_name}' not found, possible values: {', '.join(registered_code_names.keys())}",
                    file=sys.stderr,
                )
                raise ValueError()
            cls, expected_params = registered_code_names[code_name]
            kwargs = {}
            for param in params:
                assert (
                    param in expected_params
                ), f"unexpected parameter '{param}', expecting one of {', '.join(expected_params.keys())}"
                constructor = expected_params[param]
                kwargs[param] = constructor(params[param])
            self.code = cls(**kwargs)
        except Exception as e:
            print(f"[error] {e}", file=sys.stderr)
            raise e

    def __str__(self):
        return self.input

    def __call__(self):
        return self.code



import arguably
from .util import *
from . import logical_error_rate
from qec_lego_bench import noises, decoders, codes


def run():
    arguably.run()


@arguably.command
def version(name: str):
    # print(f"Goodbye, {name}!")
    print(named_kwargs_of(name))


if __name__ == "__main__":
    run()



from .util import named_kwargs_of, params_of_func_or_cls
import sys

registered_noise_names = {}


def noise_cli(*noise_names: str):

    def decorator(constructor):
        params = params_of_func_or_cls(constructor)
        for noise_name in noise_names:
            assert noise_name.isidentifier()
            noise_name = noise_name.lower()
            if noise_name in registered_noise_names:
                print(
                    f"[warning] noise name {noise_name} already registered",
                    file=sys.stderr,
                )
            registered_noise_names[noise_name] = (constructor, params)

        return constructor

    return decorator


class NoiseCli:
    def __init__(self, input: str):
        try:
            self.input = input
            noise_name, params = named_kwargs_of(input)
            noise_name = noise_name.lower()
            if noise_name not in registered_noise_names:
                print(
                    f"[error] noise name '{noise_name}' not found, possible values: {', '.join(registered_noise_names.keys())}",
                    file=sys.stderr,
                )
                raise ValueError()
            cls, expected_params = registered_noise_names[noise_name]
            kwargs = {}
            for param in params:
                assert (
                    param in expected_params
                ), f"unexpected parameter '{param}', expecting one of {', '.join(expected_params.keys())}"
                constructor = expected_params[param]
                kwargs[param] = constructor(params[param])
            self.noise = cls(**kwargs)
        except Exception as e:
            print(f"[error] {e}", file=sys.stderr)
            raise e

    def __str__(self):
        return self.input

    def __call__(self):
        return self.noise



from qec_lego_bench.cli.decoders import decoder_cli
from typing import Optional
from dataclasses import dataclass
import pathlib
import sinter
from ldpc.sinter_decoders import SinterLsdDecoder


@decoder_cli("BPLSD", "BP_LSD")
@dataclass
class BPLSD(sinter.Decoder):

    # The maximum number of iterations for the decoding algorithm
    max_iter: int = 0

    # The belief propagation method used. Must be one of {'product_sum', 'minimum_sum'}
    bp_method: str = "ms"

    # The scaling factor used in the minimum sum method
    ms_scaling_factor: float = 0.625

    # The scheduling method used. Must be one of {'parallel', 'serial'}
    schedule: str = "parallel"

    # The number of OpenMP threads used for parallel decoding
    omp_thread_count: int = 1

    # A list of integers that specify the serial schedule order. Must be of length equal to the block length of the code
    serial_schedule_order: Optional[list[int]] = None

    lsd_order: int = 0

    def decode_via_files(
        self,
        *,
        num_shots: int,
        num_dets: int,
        num_obs: int,
        dem_path: pathlib.Path,
        dets_b8_in_path: pathlib.Path,
        obs_predictions_b8_out_path: pathlib.Path,
        tmp_dir: pathlib.Path,
    ) -> None:
        decoder = SinterLsdDecoder(
            max_iter=self.max_iter,
            bp_method=self.bp_method,
            ms_scaling_factor=self.ms_scaling_factor,
            schedule=self.schedule,
            omp_thread_count=self.omp_thread_count,
            serial_schedule_order=self.serial_schedule_order,
            lsd_order=self.lsd_order,
        )
        return decoder.decode_via_files(
            num_shots=num_shots,
            num_dets=num_dets,
            num_obs=num_obs,
            dem_path=dem_path,
            dets_b8_in_path=dets_b8_in_path,
            obs_predictions_b8_out_path=obs_predictions_b8_out_path,
            tmp_dir=tmp_dir,
        )



from qec_lego_bench.codes.stabilizer_codes.css_codes.bb_code import BBCode


def test_bb_code_construction():
    code = BBCode()


import stim

tableau = stim.Tableau.from_conjugated_generators(
    xs=[
        stim.PauliString("+YZ__"),
        stim.PauliString("-Y_XY"),
        stim.PauliString("+___Y"),
        stim.PauliString("+YZX_"),
    ],
    zs=[
        stim.PauliString("+XZYY"),
        stim.PauliString("-XYX_"),
        stim.PauliString("-ZXXZ"),
        stim.PauliString("+XXZ_"),
    ],
)
print(tableau)



from qec_lego_bench.codes.stabilizer_codes.css_codes import CSSCode
from dataclasses import dataclass
from galois import GF2
import numpy as np
from galois.typing import ArrayLike
from qec_lego_bench.cli.codes import code_cli


BBExpr = tuple[int, int, int]


@dataclass
class BBCode(CSSCode):
    nkd: tuple[int, int, int]
    l: int
    m: int
    a: BBExpr
    b: BBExpr
    unsure_d: bool = False

    @property
    def n(self) -> int:
        return self.nkd[0]

    def __post_init__(self) -> None:
        # sanity check
        n, k, d = self.nkd
        assert n == 2 * self.l * self.m
        for bb_expr in [self.a, self.b]:
            for exp in bb_expr:
                assert exp >= 0
        # construct the matrices
        lm = self.l * self.m

        self.matrix_A: ArrayLike = GF2(np.zeros((lm, lm), dtype=np.uint8))
        self.matrix_A += x_matrix(self.l, self.m, self.a[0])
        self.matrix_A += y_matrix(self.l, self.m, self.a[1])
        self.matrix_A += y_matrix(self.l, self.m, self.a[2])

        self.matrix_B: ArrayLike = GF2(np.zeros((lm, lm), dtype=np.uint8))
        self.matrix_B += y_matrix(self.l, self.m, self.b[0])
        self.matrix_B += x_matrix(self.l, self.m, self.b[1])
        self.matrix_B += x_matrix(self.l, self.m, self.b[2])

        assert (self.matrix_A @ self.matrix_B == self.matrix_B @ self.matrix_A).all()
        self._H_X = np.concatenate((self.matrix_A, self.matrix_B), axis=1)
        self._H_Z = np.concatenate((self.matrix_B.T, self.matrix_A.T), axis=1)

        super().__init__()

        # assert k value using multiple equivalent methods
        assert (
            2
            * (
                self.matrix_A.shape[1]
                - np.linalg.matrix_rank(
                    np.concatenate((self.matrix_A, self.matrix_B), axis=0)
                )
            )
            == k
        )
        assert (
            2 * lm - np.linalg.matrix_rank(self.H_X) - np.linalg.matrix_rank(self.H_Z)  # type: ignore
            == k
        )
        assert len(self.logical_operators) == k

    @property
    def H_X(self) -> ArrayLike:
        return self._H_X

    @property
    def H_Z(self) -> ArrayLike:
        return self._H_Z

    def __str__(self) -> str:
        n, k, d = self.nkd
        pow_of = lambda var, exp: (
            "1" if exp == 0 else var if exp == 1 else f"{var}^{exp}"
        )
        str_A = (
            f"{pow_of('x',self.a[0])}+{pow_of('y',self.a[1])}+{pow_of('y',self.a[2])}"
        )
        str_B = (
            f"{pow_of('y',self.b[0])}+{pow_of('x',self.b[1])}+{pow_of('x',self.b[2])}"
        )
        d_str = ("\\le" if self.unsure_d else "") + str(d)
        return f"[[{n}, {k}, {d_str}]]: l={self.l}, m={self.m}, A={str_A}, B={str_B}"


bb_code_params = [
    dict(
        nkd=(72, 12, 6),
        l=6,
        m=6,
        a=(3, 1, 2),
        b=(3, 1, 2),
    ),
    dict(
        nkd=(90, 8, 10),
        l=15,
        m=3,
        a=(9, 1, 2),
        b=(0, 2, 7),
    ),
    dict(
        nkd=(108, 8, 10),
        l=9,
        m=6,
        a=(3, 1, 2),
        b=(3, 1, 2),
    ),
    dict(
        nkd=(144, 12, 12),
        l=12,
        m=6,
        a=(3, 1, 2),
        b=(3, 1, 2),
    ),
    dict(
        n=(288, 12, 18),
        l=12,
        m=12,
        a=(3, 2, 7),
        b=(3, 1, 2),
    ),
]


@code_cli("BBCode", "bb")
def bb_code_cli(n: int, k: int, d: int) -> BBCode:
    for params in bb_code_params:
        if params["nkd"] == (n, k, d):
            return BBCode(**params)  # type: ignore
    raise ValueError(f"no BBCode with n={n}, k={k}, d={d}")


def x_matrix(l: int, m: int, exp: int):
    S = GF2(np.zeros((l, l), dtype=np.uint8))
    for i in range(l):
        S[i, (i + exp) % l] = 1
    return np.kron(S, GF2(np.eye(m, dtype=np.uint8)))


def y_matrix(l: int, m: int, exp: int):
    S = GF2(np.zeros((m, m), dtype=np.uint8))
    for i in range(m):
        S[i, (i + exp) % m] = 1
    return np.kron(GF2(np.eye(l, dtype=np.uint8)), S)



from .noise import Noise
from dataclasses import dataclass
from qec_lego_bench.cli.noises import noise_cli
import stim


@noise_cli("FlipNoise", "flip")
@dataclass
class FlipNoise(Noise):
    """
    add a flip noise to all the qubits in the circuit with probability p,
    when a TICK instruction is encountered.
    """

    p: float
    basis: str = "X"  # "X", "Z" or "Y"

    def __post_init__(self):
        assert self.basis in ["X", "Y", "Z"]
        assert self.p >= 0
        assert self.p <= 1

    def _add_noise(self, circuit: stim.Circuit):
        all_qubits = list(range(circuit.num_qubits))
        circuit.append(self.basis + "_ERROR", all_qubits, self.p)

    def __call__(self, circuit: stim.Circuit) -> stim.Circuit:
        noisy = stim.Circuit()
        self.add_noise_to(circuit, noisy)
        return noisy

    def add_noise_to(self, circuit: stim.Circuit, noisy: stim.Circuit):
        for op in circuit:  # type: ignore
            if op.name == "TICK":
                noisy.append(op)
                self._add_noise(noisy)
            elif op.name == "REPEAT":
                repeat = stim.Circuit()
                self.add_noise_to(op.body_copy(), repeat)
                noisy.append(
                    stim.CircuitRepeatBlock(
                        body=repeat,
                        repeat_count=op.repeat_count,
                    )
                )
            else:
                noisy.append(op)



from qec_lego_bench.cli.decoders import decoder_cli
from typing import Optional
from dataclasses import dataclass
import pathlib
import sinter
from ldpc.sinter_decoders import SinterBeliefFindDecoder


@decoder_cli("BPUF", "BP_UF")
@dataclass
class BPUF(sinter.Decoder):

    # The maximum number of iterations for the decoding algorithm
    max_iter: int = 0

    # The belief propagation method used. Must be one of {'product_sum', 'minimum_sum'}
    bp_method: str = "ms"

    # The scaling factor used in the minimum sum method
    ms_scaling_factor: float = 0.625

    # The scheduling method used. Must be one of {'parallel', 'serial'}
    schedule: str = "parallel"

    # The number of OpenMP threads used for parallel decoding
    omp_thread_count: int = 1

    # A list of integers that specify the serial schedule order. Must be of length equal to the block length of the code
    serial_schedule_order: Optional[list[int]] = None

    # The method used to solve the local decoding problem in each cluster. Choose from: 1) 'inversion' or 2) 'peeling'.
    # By default set to 'inversion'. The 'peeling' method is only suitable for LDPC codes with point like syndromes.
    # The inversion method can be applied to any parity check matrix.
    uf_method: str = "inversion"

    # Specifies the number of bits added to the cluster in each step of the UFD algorithm. If no value is provided, this is set the block length of the code.
    bits_per_step: int = 0

    def decode_via_files(
        self,
        *,
        num_shots: int,
        num_dets: int,
        num_obs: int,
        dem_path: pathlib.Path,
        dets_b8_in_path: pathlib.Path,
        obs_predictions_b8_out_path: pathlib.Path,
        tmp_dir: pathlib.Path,
    ) -> None:
        decoder = SinterBeliefFindDecoder(
            max_iter=self.max_iter,
            bp_method=self.bp_method,
            ms_scaling_factor=self.ms_scaling_factor,
            schedule=self.schedule,
            omp_thread_count=self.omp_thread_count,
            serial_schedule_order=self.serial_schedule_order,
            uf_method=self.uf_method,
            bits_per_step=self.bits_per_step,
        )
        return decoder.decode_via_files(
            num_shots=num_shots,
            num_dets=num_dets,
            num_obs=num_obs,
            dem_path=dem_path,
            dets_b8_in_path=dets_b8_in_path,
            obs_predictions_b8_out_path=obs_predictions_b8_out_path,
            tmp_dir=tmp_dir,
        )



import arguably
import sinter

from .util import *
from .codes import *
from .noises import *
from .decoders import *
from ldpc.sinter_decoders import SinterBpOsdDecoder, SinterBeliefFindDecoder


@arguably.command
def logical_error_rate(
    code_cli: CodeCli,
    *,
    noise_cli: NoiseCli = "NoNoise",  # type: ignore
    decoder_cli: DecoderCli = "mwpf",  # type: ignore
    max_shots: int = 10_000_000,
    max_errors: int = 10_000,
    num_workers: int = 1,
):
    code = code_cli()
    noise = noise_cli()
    decoder = decoder_cli()

    ideal_circuit = code.circuit
    noisy_circuit = noise(ideal_circuit)

    task = sinter.Task(
        circuit=noisy_circuit,
        collection_options=sinter.CollectionOptions(
            max_shots=max_shots, max_errors=max_errors
        ),
    )
    sinter.collect(
        num_workers=num_workers,
        tasks=[task],
        decoders=[str(decoder_cli)],
        # decoders=["mw_parity_factor"],
        # decoders=["bposd"],
        # decoders=["bposd1"],
        # decoders=["pymatching"],
        print_progress=True,
        custom_decoders={str(decoder_cli): decoder},
    )



from qec_lego_bench.cli.decoders import decoder_cli
import chromobius
from sinter._decoding._decoding_mwpf import MwpfDecoder


@decoder_cli("Chromobius", "cb")
def chromobius_decoder():
    return chromobius.sinter_decoders()["chromobius"]
